/*
 * Pista.cpp
 *
 *  Created on: 18-lug-2009
 *      Author: gio
 */

#include <stdlib.h>
#include <JTC/jtc.h>
#include <time.h>
#include <stdio.h>
#include <iostream>

using namespace::std;

// Lunghezza pista
#define K 75
// Positione del traguardo
#define T 73


/*
 *  Piccola funzione per pulire lo schermo. System dependent: funziona solo su console Linux.
 */
void clrscr()

{
	system("clear");
}


enum Direzione { sinistra, destra, alto, basso };

class Pista;

class Visualizzatore : public JTCThread
{
	Pista* p;
public:
	Visualizzatore (Pista* p)
	{
			this->p = p;
	}
	virtual void run();

};


class Pista : public JTCMonitor
{
	char pista[4][K+1];
	int numGiocatori;
    int* posizioneX;
    int* posizioneY;
    JTCThread* currentThread;
    friend class Visualizzatore;
    Visualizzatore* v;

    /*
     *  -1 = moneta libera, 0,1,2 = codice del giocatore in possesso.
     */
    int possessoMoneta;
    int winner;
    int mosseDisponibili;

    /*
     * Riempie una linea 'l' con il carattere 'c'
     */
	void fillpista(char *l, char c)
	{
		for(int i = 0; i < K; i++)
			l[i] = c;
	    l[K] = 0; // 0-terminates each string.
	}
	/*
	 *  Ritorna in dx e dy gli incrementi necessari a spostarsi nella direzione 'dir'
	 */
	inline void getDir(Direzione dir, int& dx, int& dy)
	{
		   switch (dir)
		        {
					case sinistra:
						dx = -1;
						dy = 0;
						break;
					case destra:
						dx = 1;
						dy = 0;
						break;
					case alto:
						dx = 0;
						dy = -1;
						break;
					case basso:
						dx = 0;
						dy = 1;
		         }
	}


public:
	Pista(int numGiocatori)
	{
		possessoMoneta = -1;
		mosseDisponibili = 0;
		winner = -1;

		fillpista(pista[0],'#');
		fillpista(pista[1],' ');
		fillpista(pista[2],' ');
		fillpista(pista[3],'#');
		pista[1][T] = pista[2][T] = '|';

		this->numGiocatori = numGiocatori;
		posizioneX = new int[numGiocatori];
		posizioneY = new int[numGiocatori];

		for (int i=0; i < numGiocatori; i++)
		{
			posizioneX[i] = i / 2;
			posizioneY[i] = 1 + i % 2;
			pista[ 1 + i % 2][i / 2] = 'A'+ i;
		}
		v = new Visualizzatore(this);
		v->start();
	}

    char leggi(Direzione dir, int& dist)
    {
    	JTCSynchronized sync(*this);
        if (possessoMoneta < 0) return 0;
        int X = posizioneX[possessoMoneta];
        int Y = posizioneY[possessoMoneta];
        int dx = 0;
        int dy = 0;
        getDir (dir,dx,dy);
         dist = 0;
         do
         {
        	 dist++;
        	 X += dx;
        	 Y += dy;
         }
         while ( pista[Y][X] == ' ');
         return pista[Y][X];
    }
    int lancia(int g)
    {
    	JTCSynchronized sync(*this);
    	if (winner >= 0) return -1;
        while( possessoMoneta >= 0 ){
        	if (g == 8 || g == 9)
        		cout << g << " Dorme " << endl;
         	wait();
        }
    	if (mosseDisponibili > 0) return 0;
         possessoMoneta = g;
         srand(g*time(0));
		 /*
		 *   Bonus: tiene traccia del thread corrente per evitare invocazioni illegali di 'muovi()'
		 */
         currentThread = JTCThread::currentThread();
    	return mosseDisponibili = rand() % 2 + 1;
    }
    bool muovi(Direzione dir)
    {
    	JTCSynchronized sync(*this);
		/*
		*  Controlla che il thread corrente corrisponda all'ultimo che ha invocato 'lancia()' con successo.
		*/
    	if (mosseDisponibili == 0 || possessoMoneta < 0 || currentThread != JTCThread::currentThread()) return false;

    	bool successo = true;

        int dx;
        int dy;
        getDir(dir,dx,dy);

        int X = posizioneX[possessoMoneta];
        int Y = posizioneY[possessoMoneta];

        if (pista[Y+dy][X+dx] != ' ')
        {
           if (pista[Y+dy][X+dx] == '|')
           {
        	   winner = possessoMoneta;
           }
           successo = false;
        }
	   else
		{
			pista[Y+dy][X+dx] = pista[Y][X];
			pista[Y][X] = ' ';
			posizioneX[possessoMoneta] = X + dx;
			posizioneY[possessoMoneta] = Y + dy;
		}

        if (--mosseDisponibili == 0) {
            notifyAll();
			possessoMoneta = -1;
        }
        return successo;
    }
    int vincitore()
    {
    	JTCSynchronized sync(*this);
    	return winner;
    }
    void visualizza()
    {
    	clrscr();
    	JTCSynchronized sync(*this);
    	for (int i = 0; i < 4; i++)
    	{
    		cout << pista[i] << endl;
    	}
    	cout << "Gioca:" << possessoMoneta << endl;
    	cout << "Mosse:" << mosseDisponibili << endl;
    	if (winner >= 0)
    		cout << "Winner =" << 'A' + winner << endl;
    }
};

void Visualizzatore::run()
{
	while(p->vincitore() < 0)
	{
		sleep(100);
		p->visualizza();
	}
	// Ultima visualizzazione.
	p->visualizza();
	//getch();
}

class Automobile : public JTCThread
{
	int nt;
	Pista* p;

	Direzione pensa()
	{
		int dist;
		char c;
		c = p->leggi(destra,dist);
		// Se vedo il traguardo di fronte a me, o comunque c' spazio, vado a destra
		if (  c == '|' || dist > 1)
			return destra;
		// Cerco altrimenti di cambiare corsia
		c = p->leggi(alto,dist);
		if (dist > 1)
			return alto;
		c = p->leggi(basso,dist);
		if (dist > 1)
			return basso;
		// In assenza di alternative, viene compiuta una mossa destinata al fallimento.
		return destra;
	}

public:
	Automobile(int n, Pista* p)
	{
		nt = n;
		this->p = p;
	}
	virtual void run()
	{
		int mosse;
		while( (mosse = p->lancia(nt) ) >= 0)
		{
			while(mosse-- > 0){
				p->muovi( pensa() );
			}
			sleep(1000);
		}

	}

};

#define PLAYERS 26

int main()
{
	JTCInitialize init;
	cout << "Starting Game..." << endl;
	//clrscr();
	Pista* p = new Pista(PLAYERS);
	Automobile* cars[PLAYERS];
    for(int i=0; i < PLAYERS; i++)
    {
    	cars[i] = new Automobile(i,p);
    	cars[i]->start();
    }
}
